from .deap import deap_constructor
from .dreamer import dreamer_constructor
from .seed import seed_constructor
from .amigos import amigos_constructor
from .mahnob import mahnob_constructor
from .bci2022 import bci2022_constructor